""" Visitor constructing a string representation of an AST
    @author: Marc Provost
    @organization: McGill University
    @license: GNU General Public License
    @contact: marc.provost@mail.mcgill.ca    
"""

from ASTVisitor import ASTVisitor
from utility.TypeCheckUtility import TypeCheckUtility
from utility.ColumnConvertUtility import ColumnConvertUtility
from utility.ErrorUtility import ErrorUtility
from utility.ErrorUtility import LOG
from utility.ErrorUtility import WARNING
from utility.ErrorUtility import FATAL

typeChecker = TypeCheckUtility.getInstance()
errHandler  = ErrorUtility.getInstance()
converter   = ColumnConvertUtility.getInstance()


class StrReprVisitor(ASTVisitor):
  """Visitor constructing a string representation
     of an AST.
     @group Visitors: visit*
     @sort: __init__, visit*
  """
  
  def __init__(self):
    self.__strRepr="="
    self.__parentPriority=0
      

  def getStrRepr(self):
    """ return the constructed string representation
        @rtype: string
    """
    return self.__strRepr
              

  def __numberStr(self,number):
    val=number.getValue()
    paren=number.getNumParen()
    if int(val)-val==0:  
      val=int(val)
    for parenthesis in xrange(paren):
      self.__strRepr+="("       
    self.__strRepr+=str(val)
    for parenthesis in xrange(paren):
      self.__strRepr+=")"      

    
  def __cellRefStr(self,cellRef):
      
        
    absCol=""
    absRow=""
    paren=cellRef.getNumParen()
    if cellRef.isAbsColumn():
      absCol="$"
    if cellRef.isAbsRow():
      absRow="$"

    row=""
    col=""                        
    if cellRef.getColumn()<0:
      col="#"
    else:
      col=converter.intToString(cellRef.getColumn())

    if cellRef.getRow()<0:
      row="#"
    else:
      row=str(cellRef.getRow()+1) 
                                   
    for parenthesis in xrange(paren):
      self.__strRepr+="("                 
    self.__strRepr+=absCol+col+absRow+row
    for parenthesis in xrange(paren):
      self.__strRepr+=")"     

  def __functionStr(self,function):
    args=function.getArgs()
    paren=function.getNumParen()
    name=function.getName()
    
    for parenthesis in xrange(paren):
      self.__strRepr+="(" 

    hasRangeRef=False
    from RangeRef import RangeRef
    for arg in args:
      if isinstance(arg, RangeRef):
        hasRangeRef=True
           
    if name=="sum" and (not hasRangeRef):
      from Function import Function
      if len(args)<1:
        v=ValueError("sum should have at least 1 argument", WARNING)
        errHandler.exceptionRaised(v)      
        
      for i in xrange(len(args)):
        if i!=0:
          if not isinstance(args[i], Function):
            self.__strRepr+="+"
          else:
            if args[i].getName()=="invSum":
              self.__strRepr+="-"
            else:
              self.__strRepr+="+"
        args[i].accept(self)
    elif name=="mul" and (not hasRangeRef):
      if len(args)<1:
        v=ValueError("mul should have at least 1 argument", WARNING)
        errHandler.exceptionRaised(v) 
            
      from Function import Function
      for i in xrange(len(args)):
        if i!=0:
          if not isinstance(args[i], Function):
            self.__strRepr+="*"
          else:
            if args[i].getName()=="invMul":
              self.__strRepr+="/"
            else:
              self.__strRepr+="*"
        args[i].accept(self)
    elif name=="exp" and (not hasRangeRef):
      for i in xrange(len(args)):      
        if i!=0:              
          self.__strRepr+="^"
        args[i].accept(self)
    elif (name=="invSum" or name=="invMul") and (not hasRangeRef):
      if len(args)!=1:
        v=ValueError("invSum or invMul should have exactly one argument", 
                     WARNING)
        errHandler.exceptionRaised(v)
      args[0].accept(self) 
    elif name=="unaryMinus" and (not hasRangeRef):
      if len(args)!=1:
        v=ValueError("unary minus should have exactly one argument",
                     WARNING)
        errHandler.exceptionRaised(v)
      self.__strRepr+="-"
      args[0].accept(self)
    else:                           
      self.__strRepr+=name+"("
      for i in xrange(len(args)):
        if i!=0:
          self.__strRepr+=","
        args[i].accept(self)       
      self.__strRepr+=")" 
    
    for parenthesis in xrange(paren):
      self.__strRepr+=")" 

  def __rangeRefStr(self, rangeRef):  
    #Visit first corner: (CellRef node)
    rangeRef.getFirstCorner().accept(self)
    self.__strRepr+=":"
    #Visit second corner: (CellRef node)
    rangeRef.getSecondCorner().accept(self) 
                   
                 
                                                                    
  def visitNumber(self, number):
    """ Creates the string representation of a Number (simply its value)
    """  
    self.__numberStr(number)

  def visitCellRef(self, cellRef):
    """ Creates the string representation of a CellRef (e.g., A5, $A5, $A$5)
    """   
    self.__cellRefStr(cellRef)


  def visitFunction(self, function):
    """ Creates the string representation of a function (e.g., min(4,5))
    """  
    self.__functionStr(function)

  def visitRangeRef(self, rangeRef):
    """ Creates the string representation of a RangeRef (e.g., A1:A10)
    """
    self.__rangeRefStr(rangeRef)


