""" Testsuite for the formula grammar module
    @author: Marc Provost
    @organization: McGill University
    @license: GNU General Public License
    @contact: marc.provost@mail.mcgill.ca    
"""


from formulaparser.FormulaParser import FormulaParser
from formulaparser.FormulaToAST import FormulaToAST
from ast.ASTVisitor import ASTVisitor
from ast.Number import Number
from ast.CellRef import CellRef
from ast.RangeRef import RangeRef
from ast.Function import Function

import unittest
import sys

from math import pow
from math import fabs

from utility.ErrorUtility import ErrorUtility
from utility.TypeCheckUtility import TypeCheckUtility
from utility.ColumnConvertUtility import ColumnConvertUtility

errHandler  = ErrorUtility.getInstance()
typeChecker = TypeCheckUtility.getInstance()
converter   = ColumnConvertUtility.getInstance()

errHandler.setTesting()

class ASTSimpleEvalVisitor(ASTVisitor):
    def __init__(self):
        self.answer=0
        self.stack=[]

    def visitNumber(self, node):
        self.stack.append(node.getValue())

    def visitFunction(self, node):
        for arg in node.getArgs():
          arg.accept(self)
        if node.getName() =="sum":
            answer=0
            for i in xrange(len(node.getArgs()),0,-1):
                answer+=self.stack[-i]
            self.stack=self.stack[0:-len(node.getArgs())]
            self.stack.append(answer)
        elif node.getName() == "mul":
            answer=1
            for i in xrange(len(node.getArgs()),0,-1):
                answer*=self.stack[-i]
            self.stack=self.stack[0:-len(node.getArgs())]
            self.stack.append(answer)
        elif node.getName() == "exp":
            answer=None
            for i in xrange(len(node.getArgs()),0,-1):
                if answer==None:
                    answer=self.stack[-i]
                else:
                    answer=pow(answer,self.stack[-i])
            self.stack=self.stack[0:-len(node.getArgs())]
            self.stack.append(answer)
        elif node.getName() == "invSum":
            self.stack.append(self.stack.pop()*-1)
        elif node.getName() == "unaryMinus":
            self.stack.append(self.stack.pop()*-1)
        elif node.getName() == "invMul":
            self.stack.append(1/self.stack.pop())
        else:
            raise ValueError, "Unexpected function name: "+node.getName()

    def visitCellRef(self, astNode):
        raise NotImplementedError, "Unexpected AST node, CellRef"

    def visitRangeRef(self, astNode):
        raise NotImplementedError, "Unexpected AST node, RangeRef"

class ParserTestSuite(unittest.TestCase):
    """Testing various patterns that should be recognized
    """

    simpleExprSamples = [("=-((4))+3+7+2-1;", 7),
                         ("=-4;", -4.0),
                         ("=4.324;", 4.324),
                         ("=(4);",4),
                         ("=4+3;", 7),
                         ("=4-3;",1),
                         ("=(4-3);",1),
                         ("=4*3;",12),
                         ("=12/3;",4),
                         ("=5^2;",25),
                         ("=4+3-1+2*5;",16),
                         ("=(4+3-1+2)*5;",40),
                         ("=(4+3)*(2+1);", 21),
                         ("=(4-2)*2+(3*2);",10),
                         ("=5+5^2;", 30),
                         ("=(5+5)^2;",100),
                         ("=5^2+3*3;", 34),
                         ("=(2+3)^2;", 25),
                         ("=5^(2+2*2);", 15625),
                         ("=(5-1)^2+4+3+(3+2)^2;", 48),
                         ("=2^(3^2);", 512),
                         ("=2^(3^(2-1));", 8),
                         ("=((((4+3)*2+3+2)+3*3)*2);", 56),
                         ("=((4+(3-2)*5^2));", 29),
                         ("=-2^2^2+3;", 19),
                         ("=1.0/3.0;", 1.0/3.0),
                         ("=2*(4-10)+3^2;", -3),
                         ("=3*4^(2+3)-1^2*3;",3069),
                         ("=2*3*(1-3-(2+3)^2);",-162)
                        ]

    complexExprSamples = [("=MAX(2,3,4,5)/$A$3;",
                           Function("mul",
                             [Function("MAX",
                                [Number(2),
                                 Number(3),
                                 Number(4),
                                 Number(5)
                                ]),
                              Function("invMul",
                                [CellRef(converter.stringToInt("A"),2,1,1)
                                ])
                             ])
                          ),
                          ("=MIN(4+3,2)*2^3;",
                           Function("mul",
                             [Function("MIN",
                               [Function("sum", [Number(4), Number(3)]),
                                Number(2)
                               ]),
                              Function("exp", [Number(2), Number(3)])
                             ])
                          ),
                          ("=A5+4-2*2;",
                            Function("sum",
                              [CellRef(converter.stringToInt("A"),4,0,0),
                               Number(4),
                               Function("invSum",
                                 [Function("mul",
                                    [Number(2), Number(2)])
                                 ])
                             ])
                          ),
                          ("=$A100^2+BB$38*2+MIN(A1:A5);",
                           Function("sum",
                             [Function("exp",
                                [CellRef(converter.stringToInt("A"),99,1,0),
                                 Number(2)
                                ]),
                              Function("mul",
                                [CellRef(converter.stringToInt("BB"),37,0,1),
                                 Number(2)
                                ]),
                              Function("MIN",
                                [RangeRef(CellRef(converter.stringToInt("A"),0,0,0),
                                          CellRef(converter.stringToInt("A"),4,0,0))
                                ])
                             ])
                          )
                         ]

    formulaSuccessSamples = [("=((4));", "=((4))"),
                             ("=4+3+min(A1:A5);", "=4+3+min(A1:A5)"),
                             ("=4;", "=4"),                             
                             ("=sum(3,4,A1);", "=3+4+A1"),
                             ("=f(A1:A4,34)+g();", "=f(A1:A4,34)+g()"),
                             ("=(3+2+3)^(2+3);", "=(3+2+3)^(2+3)"),
                             ("=((23+A1+(34))+23);", "=((23+A1+(34))+23)"),
                             ("=(A1)+((sum(3,4)));", "=(A1)+((3+4))"),
                             ("=23^2*(3+2)^min(3,4);", "=23^2*(3+2)^min(3,4)"),
                             ("=3^(sum(3,4));", "=3^(3+4)"),
                             ("=r(A1:C456);","=r(A1:C456)"),
                             ("=f(A1:ZZZ9999);", "=f(A1:ZZZ9999)"),
                             ("=(((((A1)))))+5;", "=(((((A1)))))+5"),
                             ("=(((f())));", "=(((f())))"),
                             ("=((3+4^2--3));", "=((3+4^2--3))"),
                             ("=((-(4)));", "=((-(4)))"),
                             ("=f(3,4,3+4+(5));", "=f(3,4,3+4+(5))"),
                             ("=f(ZZZ9999:ZZZ9999)+g(34^3+(2*3)-((-3)));", "=f(ZZZ9999:ZZZ9999)+g(34^3+(2*3)-((-3)))"),
                             ("=3.45+((3.0));","=3.45+((3))"),
                             ("=3.45*(3.23+3^2+-2);","=3.45*(3.23+3^2+-2)"),
                             ("=f(3.45,2.33,2,g(A1:A5));", "=f(3.45,2.33,2,g(A1:A5))"),
                             ("=f(sum(3,4,3));", "=f(3+4+3)"),
                             ("=(3/(4))+3*3;", "=(3/(4))+3*3")
                            ]

    formulaFailureSamples = ["=",
                             "34",
                             "=3+4",
                             "=A1:A5;",
                             "=f(A1:A5+4);",
                             "=--5;",
                             "=zxy;",
                             "=AA;",
                             "=56B;",
                             "=A+B;",
                             "=5!;",
                             "=34+2z;",
                             "=(3));",
                             "=f(4));",
                             "=((4+5);",  
                             "=f(3,4,);",
                             "=f(A1:A2,3,A2:A4,);",
                             "=5^^4;",
                             "=3++3;",
                             "=2**3;",
                             "=f(A1:);",
                             "=A23+3,34;"
                            ]

    def testASTSuccess(self):
      """Testing AST for success
      """
      formulaToAST=FormulaToAST()
      for sample in self.formulaSuccessSamples:
        stringToParse=sample[0]
        expectedStrRepr=sample[1]
        astTuple=formulaToAST.parseToAST(stringToParse)
        self.failIf(not astTuple[0],
                    "The parser did not parse "+stringToParse+\
                    "\n"+astTuple[2])
        ast=astTuple[1]      
        self.assertEqual(expectedStrRepr,str(ast),
                         "The parser did not return the expected ast:\n"+\
                         expectedStrRepr+" != "+str(ast))  

    def testASTSanity(self):
      """Testing AST for sanity
      """
      formulaToAST=FormulaToAST()
      for sample in self.formulaSuccessSamples:
        stringToParse=sample[0]        
        astTuple=formulaToAST.parseToAST(stringToParse)
        
        self.failIf(not astTuple[0],
                    "The parser did not parse "+stringToParse+\
                    "\n"+astTuple[2])                            
        ast=astTuple[1] 
          
        strReprToReparse=str(ast)+";"
        astTupleSanity=formulaToAST.parseToAST(strReprToReparse)
        
        self.failIf(not astTupleSanity[0],
                    "The parser did not parse "+strReprToReparse+\
                    "\n"+astTupleSanity[2]) 

        astSanity=astTupleSanity[1]
                      
        self.failIf(not ast.equals(astSanity),
                    "When reparsed, the string repr did not give"+\
                    " the original AST")

    def testASTFailure(self):
      """Testing AST for failure
      """
      formulaToAST=FormulaToAST()
      for sample in self.formulaFailureSamples:
        astTuple=formulaToAST.parseToAST(sample)
        self.failIf(astTuple[0],
                    "An invalid formula was accepted and parsed: "+\
                    sample)                                                                                                                            
                                                                                                                  
    def testASTSimpleExpressionSuccess(self):
        """Testing the AST by evaluating simple arithmetic expressions
        """
        formulaToAST=FormulaToAST()
        for sample in self.simpleExprSamples:
            t=formulaToAST.parseToAST(sample[0])[1]
            v=ASTSimpleEvalVisitor()
            t.accept(v)
            ans=v.stack[0]
            self.failIf(fabs(ans-sample[1])>0.00001,
                             str(sample[1]) + " should equal " + str(ans))

    def testASTComplexExpressionSuccess(self):
        """Testing that a formula returns the expected AST
        """
        formulaToAST=FormulaToAST()
        for sample in self.complexExprSamples:
            s1=formulaToAST.parseToAST(sample[0])[1]
            s2=sample[1]
            self.failIf(not s1.equals(s2),
                        "A sample did not return the expected AST")




if __name__ == "__main__":
    unittest.main()







